require "custom_redis/version"
require "redis"

class CustomRedis

  attr_accessor :host
  attr_accessor :port
  attr_reader :redis

  @methods_key
  @type_methods
  @type

  def initialize(host = 'localhost', port = 6379)
    @host = host
    @port = port

    $custom_redis = Redis.new(:host => @host, :port => @port)
  end

  # It returns a string like model:id:method1_method2_method3
  def self.my_redis_key(model, string)
    "#{model.class.to_s.downcase}:#{model.id.to_s}:#{string}"
  end

  # It delete in Redis a key-value as { :"model:id:method1_method2_method3" => value }
  def self.cr_delete(object, custom_key)
    $custom_redis.del(my_redis_key(object, custom_key.to_s))
  end

  # It delete all Redis key-values which match with the search
  def self.cr_pflush(pattern)
    list = $custom_redis.keys(pattern)
    cr_dlist(list)
  end

  def self.cr_flush(object, methods)
    @methods_key = Array(methods).join('_')
    list = $custom_redis.keys(my_redis_key(object,@methods_key))
    cr_dlist(list)
  end



  # It evals if a key-value exists in Redis Server, being inserted if It doesn't exist
  # It returns the value after executing all methods
  # TODO: Allow to send a method which several params as a hash on *methods

  def self.cr_eval(object, type = :string, *methods)

    @methods_key = Array(methods).join('_')

    cr_evalType(type)

    case @type
      when :string
        get_method = :get
        set_method = :set
      when :set
        get_method = :smembers
        set_method = :sadd
      when :hash
        get_method = :hash
        set_method = :set
    end

    # It returns a hash with a key or an array of keys
    if (value = cr_get(object, @methods_key, get_method)) && value.present?
      value
    else
      if methods.present?
        value = object.public_send(methods[0])
        for i in 1..methods.size-1
          value = value.public_send(methods[i])
        end
      end

      # Iterate each element of a possible result for all methods of cr_eval and execute the each block of methods
      if @type == :set
        cr_iterate_set(object, value, set_method)
      else
        cr_set(object, value, set_method)
      end
    end
  end

  class << self
    private

    # It set in Redis a key-value as { :"model:id:method1_method2_method3" => value }
    def cr_set(object, value, redis_method)
      redis_key = my_redis_key(object, @methods_key)
      $custom_redis.public_send( redis_method, redis_key, value)
      value
    end

    # It get in Redis a string or list of string (depends on get method)
    def cr_get(object, custom_key, redis_method)
      $custom_redis.public_send(redis_method, my_redis_key(object, custom_key.to_s))
    end

    # It set the value in Redis within a set
    def cr_sadd(object, custom_key, value)
      $custom_redis.public_send(redis_method, my_redis_key(object, custom_key.to_s), value)
    end

    # It returns all members of a set in Redis
    def cr_smembers(object,key)
      $custom_redis.smembers(my_redis_key(object, key.to_s))
    end

    # It deletes a list of element
    def cr_dlist(list)
      list.each do |element|
        $custom_redis.del(element)
      end
    end

    # Set error messages for each public method
    def evalException(args = [])
      args.each do |hash|
        hash.map { |type, attr|
          raiseException(type, attr)
        }
      end
    end

    # Check if the attr is an active record to evaluate a possible exception to throw
    def raiseException(type, attr = nil)
      valid = ((type == "field_required") && !attr.nil?) ||
          ((type == "id_required") && !attr.nil? && attr.class.superclass.to_s.include?("ActiveRecord"))
      attr = (!attr.nil? && attr.is_a?(String)) ? attr : attr.class.to_s

      raise Exception, to_s + '::' + (attr + type.gsub('_',' ')) if !valid
    end

    # Eval the type of cr_eval method and a possible block to execute
    def cr_evalType(type)
      if type.is_a?(Hash)
        type_key = type.keys.try(:first)
        @type_methods = type.try(:[],type_key)
        @methods_key += '_' + Array(@type_methods).join('_')
      end
      @type = type_key.present? ? type_key : type
    end

    # Return an array of values for the block of methods executed
    def cr_iterate_set(object, value, set_method)
      members = []
      if !value.present?
        cr_set(object, "", set_method)
      end
      Array(value).each do |element|
        result = element
        Array(@type_methods).each do |method|
          result = result.public_send(method)
        end
        cr_set(object, result, set_method)
        members << result
      end

      members
    end

  end

end
